#include "lmath_spectrum.h"
#include "lmath_internal.h"
#include <stdlib.h>


t_lmath_errs lmath_spectrum_find_peak_freq(const double *amp_spectrum, int size,
                                           double df,
                                           const t_lmath_window_params *winpar,
                                           double search_freq,
                                           double *peak_freq, double *peak_pwr) {
    int i, fnd_i = 0;
    double fnd_pwr = 0;
    t_lmath_errs err = LMATH_ERR_OK;
    int freq_span = 0;

    if ((amp_spectrum==NULL) || (winpar == NULL) ||  ((peak_freq==NULL) && (peak_pwr==NULL))) {
        err = LMATH_ERR_INVALID_PARAMS;
    } else {
        freq_span = winpar->freq_span;
        if ((df <= 0) || (freq_span < 1)) {
            err = LMATH_ERR_INVALID_PARAMS;
        }
    }

    if (!err) {
        /* если не задана частота, в окрестности которой нужно найти требуемую,
         * то ищем во всем спектре по максимальной мощьности (исключая пост. составляющую) */
        if (search_freq <= 0) {
            for (fnd_i=-1,i= freq_span/2 + freq_span/2; i < (size - freq_span/2); i++) {
                double pwr=0;
                int j;
                for (j=i-freq_span/2; j <= (i+freq_span/2); j++) {
                    pwr+=(amp_spectrum[j]*amp_spectrum[j]);
                }

                if (pwr > fnd_pwr) {
                    fnd_pwr = pwr;
                    fnd_i = i;
                }
            }

            if (fnd_i < 0)
                err = LMATH_ERR_INTERNAL_CALC;
        } else {
            /* иначе рассчитываем мощность только в окрестности указанной частоты */
            int j;
            fnd_i = (int)(search_freq/df + 0.5);
            if (fnd_i >= (size - freq_span/2))
                fnd_i = size - freq_span/2 - 1;
            if (fnd_i < freq_span)
                fnd_i = freq_span;

            for (fnd_pwr=0, j=fnd_i-freq_span/2; j <= (fnd_i+freq_span/2); j++) {
                fnd_pwr+=(amp_spectrum[j]*amp_spectrum[j]);
            }
        }

    }

    if (!err) {
        if (peak_freq!=NULL) {
            double fnd_freq=0;
            int j;
            /* поиск частоты сигнала */
            for (j= fnd_i-freq_span/2; j <= (fnd_i+freq_span/2); j++) {
                fnd_freq+=(amp_spectrum[j]*amp_spectrum[j])*j*df;
            }
            fnd_freq/=fnd_pwr;

            *peak_freq = fnd_freq;
        }

        if (peak_pwr!=NULL) {
            *peak_pwr = fnd_pwr/winpar->enbw;
        }
    }

    return err;
}





static void f_add_spectrum_bin(t_lmath_sepctral_component *comp, double amp) {
    comp->maxBinValue = MAX(comp->maxBinValue, amp);
    comp->power += (amp*amp);
}

t_lmath_errs lmath_spectrum_calc_signoise(const double *amp_spectrum, int size,
                                          double df, const t_lmath_window_params *winpar,
                                          double peak_freq, int harm_cnt,
                                          unsigned flags, t_lmath_signoise_result *result)
{
    t_lmath_errs err = LMATH_ERR_OK;

    if ((amp_spectrum==NULL) || (winpar == NULL) ||  (result==NULL)) {
        err = LMATH_ERR_INVALID_PARAMS;
    } else {
        int i;
        int ipeak = (int)(peak_freq/df + 0.5);
        int freq_span = winpar->freq_span;
        double enbw = winpar->enbw;
        result->signal.power = result->noise.power = result->harms.power = 0;
        result->signal.maxBinValue = result->noise.maxBinValue = result->harms.maxBinValue = -100000000;
        result->params.SNR = result->params.THD = result->params.SINAD =
                result->params.SFDR = result->params.ENOB = 0;

        if (ipeak < (freq_span / 2)) {
            err = LMATH_ERR_INVALID_PARAMS;
        } else {
            for(i=freq_span/2; i<size;i++) {
                int nearest_peak_idx = (i+(freq_span / 2))/ipeak;
                int nearest_peak_idx_last = (nearest_peak_idx*ipeak)+(freq_span / 2);
                int nearest_peak_idx_first = (nearest_peak_idx*ipeak)-(freq_span / 2);

                double amp = amp_spectrum[i];

                // используется только первые harm_cnt гармоник
                if ((i>=nearest_peak_idx_first) && (i<=nearest_peak_idx_last) &&
                        (nearest_peak_idx>0) && (nearest_peak_idx<harm_cnt)) {
                    if(nearest_peak_idx > 1) {
                        f_add_spectrum_bin(&result->harms, amp);
                    } else {
                        f_add_spectrum_bin(&result->signal, amp);
                    }
                } else {
                    f_add_spectrum_bin(&result->noise, amp);
                }
            }

            result->signal.power /= enbw;
            result->noise.power /= enbw;
            result->harms.power /= enbw;

            if(result->signal.power > 0) {
                if (flags & LMATH_SIGNOISE_FLAGS_CALC_SFDR) {
                    double DynamicRange = result->signal.maxBinValue/(MAX(result->harms.maxBinValue,
                                                                     result->noise.maxBinValue));
                    if (DynamicRange > 0) {
                        result->params.SFDR = 20 * log10(DynamicRange);
                    }
                }

                if (flags & LMATH_SIGNOISE_FLAGS_CALC_SNR) {
                    if (result->noise.power > 0) {
                        double SNRPower =  result->signal.power / result->noise.power;
                        if((SNRPower > 0)) {
                            result->params.SNR = 10 * log10(SNRPower);
                        }
                    }
                }

                if (flags & LMATH_SIGNOISE_FLAGS_CALC_THD) {
                    if(result->harms.power > 0) {
                        double THDPower = result->harms.power/result->signal.power;
                        if (THDPower>0) {
                            result->params.THD = 10 * log10(THDPower);
                        }
                    }
                }

                if ((flags & LMATH_SIGNOISE_FLAGS_CALC_SINAD) ||
                    (flags & LMATH_SIGNOISE_FLAGS_CALC_ENOB)) {
                    if (result->noise.power > 0) {
                        double SINADPower = result->signal.power/result->noise.power;
                        if (SINADPower>0) {
                            result->params.SINAD = 10*log10(SINADPower);
                            result->params.ENOB = (result->params.SINAD - 1.76) / 6.02;
                        }
                    }
                }
            }
        }
    }
    return err;
}
