#ifndef LMATH_SPECTRUM_H
#define LMATH_SPECTRUM_H

#include "lmath_window.h"

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************//**
    Поиск значения частоты с максимальной мощностью и мощности сигнала на этой
    частоте по амплитудному спектру.

    @param[in]  amp_spectrum    Амплитудный спектр входного сигнала
    @param[in]  size            Количество отсчетов в спектре
    @param[in]  df              Изменение частоты между соседними отсчетами
                                  спектра в Гц
    @param[in]  winpar          Указатель на структуру с параметрами
                                   примененного перед вычислением спектра окна.
    @param[in]  search_freq     Если значение больше нуля, то ищется частота в
                                  окрестности указанной, иначе - во всей частотной
                                  области (кроме постоянной составляющей)
    @param[out] peak_freq       В этой переменной возвращается рассчитанное значение
                                  частоты, соответствующей максимальной мощности.
                                  Может быть NULL.
    @param[out] peak_pwr        В этой переменной возвращается рассчитанное значение
                                  мощности сигнала на найденной частоте.
                                  Может быть NULL.
    @return                     Код ошибки
*******************************************************************************/
t_lmath_errs lmath_spectrum_find_peak_freq(const double *amp_spectrum, int size, double df,
                                           const t_lmath_window_params *winpar,
                                           double search_freq,
                                           double *peak_freq, double *peak_pwr);






/** Флаги, определяющие, какие результирующие параметры будут рассчитаны
 * при вызове lmath_spectrum_calc_signoise() */
typedef  enum {
    LMATH_SIGNOISE_FLAGS_CALC_SNR   = 1 << 0, /**< Признак необходимости расчета SNR */
    LMATH_SIGNOISE_FLAGS_CALC_THD   = 1 << 1, /**< Признак необходимости расчета THD */
    LMATH_SIGNOISE_FLAGS_CALC_SINAD = 1 << 2, /**< Признак необходимости расчета SINAD */
    LMATH_SIGNOISE_FLAGS_CALC_SFDR  = 1 << 3, /**< Признак необходимости расчета SFDR */
    LMATH_SIGNOISE_FLAGS_CALC_ENOB  = 1 << 4, /**< Признак необходимости расчета ENOB */
} t_lmath_signoise_flags;


/** Рассчитанные значения, по одному из спектральных компонент сигнала */
typedef  struct {
    double power; /**< Мощность */
    double maxBinValue; /**< Максимальное значение отсчета в спектре */
}  t_lmath_sepctral_component;


/** Результаты выполнения функции lmath_spectrum_calc_signoise() */
typedef struct {
    t_lmath_sepctral_component signal; /**< Значение спектральной составляющей,
                                            относящейся к рассматриваемому сигналу */
    t_lmath_sepctral_component harms; /**< Значение спектральной составляющей,
                                            относящейся к гармоникам рассматриваемого
                                            сигнала */
    t_lmath_sepctral_component noise; /**< Значение спектральной составляющей,
                                            относящейся к шуму (все вне сигнала и
                                            гармоник) */

    struct {
        double SNR;  /**< Значение отношения сигнал/шум в дБ */
        double THD;  /**< Значение Total Harmonic Distortion:
                          гармоники/сигнал дБ */
        double SINAD;/**< Значение signal-to-noise-and-distortion
                          ratio: сигнал/(шум+гармоники) дБ. */
        double SFDR; /**< Значение spurious free dynamic range:
                          макс. отсчет сигнала/макс. отсчет вне сигнала дБ */
        double ENOB; /**< Значение количества эффективных бит:
                          (SINAD - 1.76) / 6.02 */
    } params;
} t_lmath_signoise_result;


/***************************************************************************//**
    Вычисление значений сигнала/гармоник/шума в спектре и расчет производных
    параметров от данных составляющих.

    @param[in]  amp_spectrum    Амплитудный спектр входного сигнала
    @param[in]  size            Количество отсчетов в спектре
    @param[in]  df              Изменение частоты между соседними отсчетами
                                  спектра в Гц
    @param[in]  winpar          Указатель на структуру с параметрами
                                   примененного перед вычислением спектра окна.
    @param[in]  peak_freq       Частота основного сигнала. Должна соответствовать
                                   отсчету с номером не меньше чем freq_span/2.
    @param[in]  harm_cnt        Количество гармоник (включая основную), используемых
                                    при расчете.
    @param[in]  flags           Набор флагов из #t_lmath_signoise_flags,
                                    определяющих, какие параметры рассчитывать.
    @param[out]  result         При успешном выполнении в данную структуру сохраняются
                                    рассчитанные значения.


    @return                     Код ошибки
*******************************************************************************/
t_lmath_errs lmath_spectrum_calc_signoise(const double *amp_spectrum, int size,  double df,
                                            const t_lmath_window_params *winpar,
                                            double peak_freq, int harm_cnt,
                                            unsigned flags,
                                            t_lmath_signoise_result *result);



#ifdef __cplusplus
}
#endif

#endif // LMATH_SPECTRUM_H
